/* Written by Simon Lacoste-Julien on 2002/07/29 */
/* This is a test for Python Extension and fortran linking
I'm implementing the Euler solver used in the 522 Simulator, but using fortran code
Interface of use:
>>> import fEuler
>>> newx = fEuler.step(f,x,t,delta_t) # almost as before!
The difference is that f is no more an array of (derivative) functions, but it
is called as f(i,x,t) where i will be the index of the derivative required...
x is initial array of values,
t is time, and delta_t is step length (both are double).  And the return value
is a simple double.
*/

#include <Python.h>

/* SUBROUTINE EULER(N,fun(N,i,x,t),x,t,dt,newx) imported from Fortran: */
extern void euler_(int* N, void (*fun)(int* N1, int* i, double x1[], double* t1, double* res), double x[],
           double* t, double* dt, double newx[]);
/* notes: 1) euler_ is in lower-case and have added '_' for the name compatibility with Fortran
          2) all the arguments are passed as pointers, since Fortran pass arguments by reference
      and not by value... */

/***** Callback function interface between C and python for the derivative function: ************/
/* CALL-BACK POINTER: */
static PyObject* funPython = NULL;  /* this will hold the function passed by Python; it is GLOBAL! */

/* CALL-BACK FUNCTION WRAPPER: */
static void callBack(int* N, int* i, double x[], double* t, double* res) {
  PyObject* arglist = NULL;
  PyObject* x_python = NULL;
  int j;

  /* arguments translation: */
  x_python = PyList_New(*N);
  Py_INCREF(x_python);
  for(j = 0; j < (*N); j++) {
    PyList_SetItem(x_python, j, Py_BuildValue("d", x[j]));
  }

  arglist = Py_BuildValue("(i,O,d)", *i, x_python, *t);    /* building (i,x,t)  NOTE: x is a vector! */
  Py_INCREF(arglist);
  *res = PyFloat_AsDouble(PyEval_CallObject(funPython, arglist));    /* *res = (double) funPython(i,x,t) */

  /* memory cleaning: */
  Py_DECREF(arglist);
  Py_DECREF(x_python);

  return;
}


/***** 2) Module method definition ********/
static PyObject*
step(PyObject* self, PyObject* args) {

  PyObject* x_python = NULL;        /* will hold array x in Python */
  PyObject* newx = NULL;          /* will hold result (array) in Python */
  PyObject* xObject = NULL;       /* will hold Python x[i] */
  double* x_array;                /* array to hold values of x in C format */
  double* workArray;              /* array to hold new values of x in Fortran call  */
  int length;             /* size of x */
  int i;                  /* array index */
  double t, dt;

  /* getting arguments from python: */
  if (PyArg_ParseTuple(args, "OOdd",
               &funPython, &x_python, &t, &dt) == 0) {       /* converts Python -> C */
    /* setting error message: */
    PyErr_SetString(PyExc_RuntimeError, "Wasn't able to parse args for Euler.step fct.");
    return NULL;        /* when returning null, you raise an exception */
  } else {
    Py_INCREF(x_python);
    Py_INCREF(funPython);
    if(PyList_Check(x_python) == 0) {
      PyErr_SetString(PyExc_TypeError, "Error: x wasn't a list!");
      goto error;
    }

    /************ converting Python to C/Fortran: **************/
    length = PySequence_Length(x_python);
    x_array = calloc(length, sizeof(double));    /* x array in C */
    workArray = calloc(length, sizeof(double));  /* to be used by Fortran code */

    /* initializing x_array: */
    for(i = 0; i < length; i++) {
      Py_XDECREF(xObject);
      xObject = PyList_GetItem(x_python, i);
      Py_INCREF(xObject);
      x_array[i] = PyFloat_AsDouble(xObject);
    }

    /* CALLING Fortran code: */
    euler_(&length, callBack, x_array, &t, &dt, workArray);

    /************* translating result to Python *****************/

    newx = PyList_New(length);              /* initializing the return array */
    Py_INCREF(newx);
    for(i = 0; i < length; i++) {
      PyList_SetItem(newx, i, Py_BuildValue("d", workArray[i]));
    }

    /* cleaning memory: */
    Py_XDECREF(x_python);
    /* not doing it on newx since we're returning it */
    Py_XDECREF(xObject);
    Py_XDECREF(funPython);
    free(x_array);
    free(workArray);

    return newx;
  }
 error:
  /* cleaning memory: */
  Py_XDECREF(x_python);
  Py_XDECREF(newx);
  Py_XDECREF(xObject);
  Py_XDECREF(funPython);

  return NULL;        /* this raises exception */
}

/***** 3) Registration table  **************/
/* this gives the association between the function NAMES (to be used
   in Python namespace) to function pointers (addresses) giving the C implementation */

static struct PyMethodDef fEuler_methods[] = {
  {"step", step, METH_VARARGS, "Call Fortran Euler Solver."},
  /* method name, C func ptr., method of calling fct, method doc */
  {NULL, NULL, 0, NULL}                    /* end of table marker */
};

/***** 4) Initialization function for the module *****/
/* this is called the first time the C module is imported in Python;
   it builds the Module's attribute dictionary from the registration
   table and create an entry for the C module on the sys.module table */
/* name of fct matters: should be init<moduleName> */
void initfEuler() {
  (void) Py_InitModule("fEuler", fEuler_methods);   /* module name, table ptr */
}
